package jgame;
import java.awt.*;
import java.awt.geom.*;
import java.awt.font.*;
import java.applet.*;
import java.awt.event.*;
import javax.swing.ListCellRenderer;
import javax.swing.JList;
import java.net.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;
/** Contains the main functionality of the game engine.  Subclass it to write
 * your own game.  The engine can be run as both an applet and an application.
 * To run as an application, have your main() construct an instance of your
 * subclass. Then call initEngine(), which will open a window, and initialise.
 * To run as an applet, ensure that your subclass has a parameterless
 * constructor which enables the applet viewer to create it as an applet.
 * Within this constructor, you call initEngineApplet() to initialise.

 * <P> Upon initialisation, the engine will create a new thread in which all
 * game action will run.  First it calls initGame() in your subclass.  Here,
 * you can define any initialisation code, such as variable initialisation and
 * image loading.  Then, the engine will start producing frames.  Each frame
 * it will call doFrame() in your subclass, where you can call engine
 * functions to move the game objects, check collisions, and everything else
 * you want to do in a frame.  Then, it will draw the current frame on the
 * screen, after which it calls paintFrame(), where you can do any customised
 * drawing on the frame, such as status messages.  The engine enables you to
 * specify <i>states</i> that the game may be in, say, the title screen state
 * or the in-game state.  When a state is set, the engine will try to call
 * additional doFrame<i>state</i>() and paintFrame<i>state</i>() methods,
 * where you can handle the game action for that specific state. Multple
 * states may be set at once, resulting in multiple method calls each frame.

 * <P> The engine manages a list of sprite objects, and a matrix of
 * equal-sized tiles. Its primary way to draw graphics is images. Images may
 * be loaded from Gifs, JPegs, or PNGs (PNGs not in java 1.2), which may
 * contain either single images, or regularly spaced arrays or matrices of
 * images (i.e. sprite sheets, we will call them image maps).  An image is
 * identified by a unique name, and is defined bythe image file it comes from,
 * the index within that file in case the file is an image map, and any flip
 * and rotate actions that are to be be done on the file image to arrive at
 * the final image.  Animations may be defined in terms of sequences of
 * images.  Image maps, images, and animations can be defined by calling resp.
 * defineImageMap, defineImage, or defineAnimation, or the definitions can be
 * loaded from a text-file table file using defineImages().

 * <P>
 * Objects are of the class JGObject; see this class for more information.
 * Objects are identified within the engine by unique String identifiers.
 * Creating a new object with the same identifier as an old object will
 * replace the old one.  Objects are drawn on the screen in lexical order.
 * The main methods for handling objects are moveObjects(), checkCollision(),
 * and checkBGCollision().

 * <P>
 * Tiles can be used to define a background that the sprite objects can
 * interact with, using a set of shorthands for reading, writing, and
 * colliding with tiles.  Tiles are uniquely identified by a short string
 * (1-4 characters).  Tiles may be transparent; a decorative background
 * image can be displayed behind the tiles. 

 * <P> Collision is done by assigning collision IDs (cids) to both objects and
 * tiles. A cid is basically a bit string (usually, one allocates 1 bit per
 * basic object or tile type).  Collision is done by specifying that a certain
 * set of object types should be notified of collision with a certain set of
 * object or tile types.  Such a set is specified by a bit mask, which matches
 * a cid when cid&amp;mask != 0.  JGObject defines collision methods (hit and
 * hit_bg)which can then be implemented to handle the collisions.  Objects
 * have two bounding boxes: one for object-object collision (the "regular"
 * bounding box), and one for object-tile collision (the tile bounding box).
 * The two bounding boxes are defined by the image (typically, you want these
 * bounding boxes to approximate the shape of each individual image).  Both
 * can be overridden separately. Often, you want to set the tile bbox to a
 * constant value, such as equal to the size of 1 tile.

 * <P>
 * The engine supports transparent on the fly scaling.  This means that you
 * can code your game for a specific "virtual" screen size, and have it scale
 * to any other screen size by simply supplying the desired "real" screen size
 * at startup.  Applets will scale to the size as given by the width and height
 * fields in the HTML code.  The scale factor will in no way affect the
 * behaviour of the game (except performance-wise), but the graphics may look
 * a bit blocky or jaggy, as the engine uses a simple scaling algorithm to
 * ensure good performance.

 * Currently, the engine supports keyboard input only.  Mouse input will be
 * added later.  The state of the keyboard is maintained in a keymap, which
 * can be read by getKey().

 */
public abstract class JGEngine extends Applet {

	protected static int debugflags = 0;

	public static void setBoundingBoxDebug(boolean enabled) {
		if (enabled) debugflags |= 1;
		else         debugflags &= 0xfffe;
	}

	/** Path from where files can be loaded; null means not initialised yet */
	protected String path = null;
	private Thread thread=null;
	protected JGCanvas canvas=null;
	private JGListener listener = new JGListener();
	private double fps = 25;
	private double maxframeskip = 4.0; /* max # of frames to skip  */
	private long target_time=0; /* time at which next frame should start */
	/** Should engine thread run or halt? Set by start() / stop()*/
	private boolean running=true;
	/** Engine game state */
	Vector gamestate=new Vector(10,20);
	/** indicates when engine is inside a parallel object update (moveObjects,
	 * check*Collision) */
	boolean in_parallel_upd=false;
	private Vector timers = new Vector(20,40);


	/** signals that JGame globals are set, and exit code should null globals in
	* JGObject */
	boolean is_inited=false;
	/** signals that thread should die */
	boolean is_exited=false;

	protected boolean i_am_applet=false;

	public static final int KeyUp=38;
	public static final int KeyDown=40;
	public static final int KeyLeft=37;
	public static final int KeyRight=39;
	public static final int KeyShift=16;
	public static final int KeyCtrl=17;
	public static final int KeyAlt=18;
	
	/* playfield dimensions */

	/** width/height after scaling to screen; 0 is not initialised yet */
	int width=0,height=0; 
	int nrtilesx,nrtilesy;
	int tilex,tiley;
	Frame my_win=null;

	protected Color fg_color = Color.white;
	protected Color bg_color = Color.black;
	protected Font msg_font=null;

	Hashtable animations = new Hashtable();

	protected class Animation implements Cloneable {
		/* settings;  the public ones may be manipulated freely */
		String [] frames;
		public double speed;
		public int increment=1;
		public boolean pingpong=false;
		/* state */
		int framenr=0;
		double phase=0.0;
		boolean running=true;
		public Animation (String [] frames, double speed) {
			this.frames=frames;
			this.speed=speed;
		}
		public Animation (String [] frames, double speed, int increment) {
			this.frames=frames;
			this.speed=speed;
			this.increment=increment;
		}
		public Animation (String [] frames, double speed, int increment,
		boolean pingpong) {
			this.frames=frames;
			this.speed=speed;
			this.increment=increment;
			this.pingpong=pingpong;
		}
		public void stop() { running=false; }
		public void start() { running=true; }
		public void reset() {
			framenr=0;
			phase=0.0;
		}
		public Object clone() {
			try {
				return super.clone();
			} catch (CloneNotSupportedException e) { return null; }
		}
		public Animation copy() {
			return (Animation) clone();
		}
		public String getCurrentFrame() {
			if (framenr < 0 || framenr >= frames.length) {
				return frames[0];
			} else {
				return frames[framenr];
			}
		}
		/** Does one animation step and returns current image.  Note that the
		 * function returns the frame before the state is being updated. */
		public String animate() {
			String ret = getCurrentFrame();
			if (running) {
				phase += speed;
				if (phase > 1.0) {
					phase=0.0;
					framenr += increment;
					if (!pingpong) {
						if (framenr >= frames.length) framenr -= frames.length;
						if (framenr < 0)              framenr += frames.length;
					} else {
						if (framenr >= frames.length) {
							framenr -= increment;
							increment = -increment;
						}
						if (framenr < 0) {
							framenr -= increment;
							increment = -increment;
						}
					}
				}
			}
			return ret;
		}
	}

	public void defineAnimation (String id,
	String [] frames, double speed) {
		animations.put(id, new Animation(frames,speed));
	}

	public void defineAnimation (String id,
	String [] frames, double speed, int increment) {
		animations.put(id, new Animation(frames,speed,increment));
	}

	public void defineAnimation (String id,
	String [] frames, double speed, int increment, boolean pingpong) {
		animations.put(id, new Animation(frames,speed,increment,pingpong));
	}


	protected Animation getAnimation(String id) {
		return (Animation) animations.get(id);
	}

	/** Construct engine, but do not initialise it yet. 
	* Call initEngine or initEngineApplet to initialise the engine. */
	public JGEngine() {}

	/** Init engine as applet; call this in your engine constructor.  Applet
	 * init() will start the game.
	 * @param nrtilesx  nr of tiles horizontally
	 * @param nrtilesy  nr of tiles vertically
	 * @param tilex  width of one tile
	 * @param tiley  height of one tile
	 * @param fgcolor pen/text colour, null for default white
	 * @param bgcolor background colour, null for default black
	 * @param msgfont font for messages and text drawing, null for default
	 */
	public void initEngineApplet(int nrtilesx,int nrtilesy, int tilex,int tiley,
	Color fgcolor, Color bgcolor, Font msgfont){
		i_am_applet=true;
		this.nrtilesx=nrtilesx;
		this.nrtilesy=nrtilesy;
		this.tilex=tilex;
		this.tiley=tiley;
		setColorsFont(fgcolor,bgcolor,msgfont);
		// we get the width/height only after init is called
		path = null;
	}

	/** Init engine as application. 
	 * @param nrtilesx  nr of tiles horizontally
	 * @param nrtilesy  nr of tiles vertically
	 * @param tilex  width of one tile
	 * @param tiley  height of one tile
	 * @param fgcolor pen/text colour, null for default white
	 * @param bgcolor background colour, null for default black
	 * @param msgfont font for messages and text drawing, null for default
	 * @param width  real screen width, 0 = use screen size
	 * @param height real screen height, 0 = use screen size */
	public void initEngine(int nrtilesx,int nrtilesy, int tilex,int tiley,
	Color fgcolor, Color bgcolor, Font msgfont,
	int width,int height) {
		i_am_applet=false;
		this.nrtilesx=nrtilesx;
		this.nrtilesy=nrtilesy;
		this.tilex=tilex;
		this.tiley=tiley;
		if (width==0) {
			Dimension scrsize = Toolkit.getDefaultToolkit().getScreenSize();
			this.width = scrsize.width;
			this.height = scrsize.height-32; // see also createWindow...
		} else {
			this.width=width;
			this.height=height;
		}
		setColorsFont(fgcolor,bgcolor,msgfont);
		path = System.getProperty("user.dir")+"/";
		createWindow();
		init();
	}

	/** Set foreground and background colour and font in one go; passing null
	 * means ignore. */
	public void setColorsFont(Color fgcolor,Color bgcolor,Font msgfont) {
		if (msgfont!=null) setFont(msgfont);
		if (fgcolor!=null) setColor(fgcolor);
		if (bgcolor!=null) setBGColor(bgcolor);
	}

	/** Get the virtual width in pixels (not the scaled screen width) */
	public int pfWidth() { return nrtilesx*tilex; }
	/** Get the virtual height in pixels (not the scaled screen height) */
	public int pfHeight() { return nrtilesy*tiley; }

	/** Get the number of tiles in X direction */
	public int pfTilesX() { return nrtilesx; }
	/** Get the number of tiles in Y direction */
	public int pfTilesY() { return nrtilesy; }

	/** Get the tile width in (virtual) pixels. */
	public int pfTileWidth()  { return tilex; }
	/** Get the tile height in (virtual) pixels. */
	public int pfTileHeight() { return tiley; }

	/** Add base URL of the current environment to the given file path; this
	 * will result in a URL that both applications and applets can load from. 
	* @param subpath  relative path, null if error */
	public URL getFullPath(String subpath) {
		try {
			return new URL(canvas.baseurl,canvas.basepath+subpath);
		} catch (MalformedURLException e) {
			System.err.println("Error: URL format error for '"+subpath+"'.");
			return null;
		}
	}

	/** Initialise engine; don't call directly.  This is supposed to be called
	 * by the applet viewer or the initer.
	 */
	public void init() {
		if (width==0) {
			// get width/height from applet dimensions
			width=getWidth();
			height=getHeight();
		}
		if (path==null) {
			URL codebase=getCodeBase();
			canvas = new JGCanvas(this, nrtilesx,nrtilesy, tilex,tiley,
				width,height,codebase,"");
		} else try {
			canvas = new JGCanvas(this, nrtilesx,nrtilesy, tilex,tiley,
				width,height,new URL("file://"),path);
		} catch (MalformedURLException e) { System.out.println(e); }
		//getAudioClip(getFullPath("sounds/mysound")).play();
		// set bg color so that the canvas's padding is in the proper color
		canvas.setBackground(bg_color);
		if (my_win!=null) my_win.setBackground(bg_color);
		// determine default font size
		msg_font = new Font("Helvetica",0,(int)(16.0/(640.0/(tilex*nrtilesx))));
		setFont(msg_font);
		setLayout(new FlowLayout(FlowLayout.LEADING,0,0));
		add(canvas);
		if (!JGObject.setEngine(this,canvas)) {
			/** yes, you see that right.  I've used a random interface with a
			 * method that allows me to pass a Graphics.  We shall move this
			 * stuff to JGCanvas later, i suppose */
			canvas.setInitPainter(new ListCellRenderer () {
				public Component getListCellRendererComponent(JList d1,
				Object value, int d2, boolean d3, boolean d4) {
					Graphics g = (Graphics) value;
					g.setFont(msg_font);
					g.setColor(bg_color);
					g.fillRect(0,0,width,height);
					g.setColor(fg_color);
					drawString(g,"JGame is already running in this VM",
						pfWidth()/2,pfHeight()/3,0);
					return null;
				} } );
			return;
		}
		is_inited=true;
		canvas.setInitPainter(new ListCellRenderer () {
			public Component getListCellRendererComponent(JList d1,
			Object value, int d2, boolean d3, boolean d4) {
				Graphics g = (Graphics) value;
				g.setFont(msg_font);
				g.setColor(bg_color);
				g.fillRect(0,0,width,height);
				g.setColor(fg_color);
				drawString(g,"Please wait, loading files .....",
					pfWidth()/2,pfHeight()/3,0);
				if (canvas.progress_message!=null) {
					drawString(g,canvas.progress_message,
							pfWidth()/2,2*pfHeight()/3,0);
				}
				return null;
			} } );
		//addKeyListener(listener); //this was here until 200608
		if (my_win!=null) {
			my_win.setVisible(true);
			my_win.validate();
		}
		// initialise keyboard handling
		canvas.addKeyListener(listener);
		canvas.requestFocus();
		thread = new Thread(new JGEngineThread());
		thread.start();
	}

	/** Override to define your own initialisations.  This method
	* is called after calling initEngine() or initEngineApplet().. */
	abstract public void initGame();

	/** Signal that the engine should start running */
	public void start() { running=true; }
	/** signal that the engine should stop running and wait */
	public void stop() { running=false; }

	/** Destroy function for deinitialising the engine properly.  In
	* particular, this is called by the applet viewer to dispose the applet. */
	public void destroy() {
		// kill game thread
		is_exited=true;
		// applets cannot interrupt threads; their threads will 
		// be destroyed for them (not always, though ...).
		if (thread!=null) {
			if (!i_am_applet) thread.interrupt();
			try {
				thread.join(2000); // give up after 2 sec
			} catch (InterruptedException e) {
				e.printStackTrace();
				// give up
			}
		}
		// remove frame??
		// close files?? that appears to be unnessecary
		// reset global variables
		if (is_inited) {
			JGObject.default_engine=null;
			JGObject.default_canvas=null;
		}
		System.out.println("JGame engine disposed.");
	}


	/** Set frame rate in frames per second, and maximum number of frames that
	 * may be skipped before displaying a frame again.
	 * @param fps  frames per second, useful range 2...80
	 * @param maxframeskip  max successive frames to skip, useful range 0..10*/
	public void setFrameRate(double fps, double maxframeskip) {
		this.fps = fps;
		this.maxframeskip = maxframeskip;
	}

	/** Set global background colour. */
	public void setBGColor(Color bgcol) {
		bg_color=bgcol;
		if (canvas!=null) canvas.setBackground(bgcol);
		if (my_win!=null) my_win.setBackground(bgcol);
	}

	/** Set image to display behind transparent tiles.  Image size must be a
	 * multiple of the tile size. Passing null turns off background image; the
	 * background colour will be used instead.
	 * @param bgimg  image name, null=turn off background image */
	public void setBGImage(String bgimg) { canvas.setBGImage(bgimg); }

	/** Set global foreground colour, used for printing text and status
	 * messages. */
	public void setColor(Color fgcol) {
		fg_color=fgcol; 
		// if we are drawing status, set the color of the graphics context
		if (buf_gfx!=null) buf_gfx.setColor(fgcol);
	}

	/** Set font for printing text and status messages. */
	public void setFont(Font msgfont) {
		if (canvas==null) {
			// set unscaled font, scale later when we define canvas
			this.msg_font=msgfont;
		} else {
			double origsize = msgfont.getSize2D();
			msg_font=msgfont.deriveFont((float)(origsize*canvas.min_scale_fac));
			//System.out.println("orig"+origsize+" scalefac"+canvas.min_scale_fac+
			//	" scaled "+msg_font);
			// if we are drawing status, set the font of the graphics context
			if (buf_gfx!=null) buf_gfx.setFont(msgfont);
		}
	}


	protected void registerTimer(JGTimer timer) { timers.add(timer); }

	private void tickTimers() {
		for (int i=timers.size()-1; i>=0; i--) {
			JGTimer timer = (JGTimer)timers.get(i);
			if (timer.tick()) {
				timers.remove(timer);
			}
		}
	}


	/** Set the game's main state.  Methods with the names
	 * doFrame&lt;state&gt; and paintFrame&lt;state&gt; will be called in
	 * addition to doFrame() and paintFrame(). */
	public void setGameState(String state) {
		gamestate.clear();
		gamestate.add(state);
	}

	/** Add the given state to the game's existing state.  The methods
	* doFrame&lt;state&gt; and paintFrame&lt;state&gt; will be called <i>in
	* addition to</i> the methods of any states already set. */
	public void addGameState(String state) {
		gamestate.add(state);
	}

	/** Set the game's main state to none.  Only doFrame() and paintFrame()
	 * will be called each frame. */
	public void clearGameState() { gamestate.clear(); }


	/** Check if game is in given state. */
	public boolean inGameState(String state) {
		for (Enumeration e=gamestate.elements(); e.hasMoreElements(); ) {
			if ( ((String)e.nextElement()).equals(state) ) return true;
		}
		return false;
	}

	/** Do some administration, call doFrame. */
	private void doFrameAll() {
		// remove any objects that were created in the main routine after the
		// last moveObjects or checkCollision
		canvas.flushRemoveList();
		canvas.flushAddList();
		tickTimers();
		canvas.flushRemoveList();
		canvas.flushAddList();
		doFrame();
		invokeGameStateMethods("doFrame");
	}

	private void invokeGameStateMethods(String prefix) {
		for (Enumeration e=gamestate.elements(); e.hasMoreElements(); ) {
			String state = (String) e.nextElement();
			try {
				Method met=getClass().getMethod(prefix+state, null);
				met.invoke(this,new Object[0]);
			} catch (NoSuchMethodException ex) {
				// ignore when method is not defined
				//System.err.println("Error: cannot find method "+prefix+state);
			} catch (Exception ex) {
				System.err.println("Exception while executing "+prefix+state);
				ex.printStackTrace();
			}
		}
	}

	/** Is called every frame.  Override to define frame action.  Default is
	 * do nothing. */
	public void doFrame() {}

	private Graphics buf_gfx=null;
	protected void paintFrame(Graphics g) {
		buf_gfx=g;
		g.setFont(msg_font);
		g.setColor(fg_color);
		paintFrame();
		invokeGameStateMethods("paintFrame");
		buf_gfx=null;
	}

	/** Is called when the engine's default screen painting is finished,
	* and custom painting actions may be carried out. Can be used to display
	* status information or special graphics.  Default is do nothing. */
	public void paintFrame() {}

	/** Get the graphics context for drawing on the buffer during a
	 * paintFrame().  Returns null when not inside paintFrame. */
	public Graphics getBufferGraphics() { return buf_gfx; }

	/** Get scale factor of real screen width wrt virtual screen width */
	public double getXScaleFactor() { return canvas.x_scale_fac; }
	/** Get scale factor of real screen height wrt virtual screen height */
	public double getYScaleFactor() { return canvas.y_scale_fac; }
	/** Get minimum of the x and y scale factors */
	public double getMinScaleFactor() { return canvas.min_scale_fac; }

	/* some convenience functions for drawing during paintFrame() */

	/** Draw image with given ID */
	public void drawImage(int x,int y,String imgname) {
		canvas.drawImage(buf_gfx,x,y,imgname);
	}

	/** draws string so that (x,y) is the topleft coordinate (align=-1), the
	 * top middle coordinate (align=0), or the top right coordinate (align=1).
	 * @param align  text alignment, -1 = left, 0=center, 1=right */
	public void drawString(String str, int x, int y, int align) {
		drawString(buf_gfx, str, x,y, align);
	}

	protected void drawString(Graphics g, String str, int x, int y, int align){
		Point p = canvas.scalePos(x,y);
		Font font = g.getFont();
		FontRenderContext fontrc = ((Graphics2D)g).getFontRenderContext();
		//Rectangle2D fontbounds = font.getMaxCharBounds(fontrc);
		//Rectangle2D stringbounds = getStringBounds(str, fontrc);
		TextLayout layout = new TextLayout(str, font, fontrc);
		Rectangle2D strbounds = layout.getBounds();
		if (align==-1) {
			g.drawString(str,
					(int)(p.x-strbounds.getMinX()),
					(int)(p.y-strbounds.getMinY()));
		} else if (align==0) {
			g.drawString(str,
					(int)(p.x-strbounds.getCenterX()),
					(int)(p.y-strbounds.getMinY()));
		} else {
			g.drawString(str,
					(int)(p.x-strbounds.getMaxX()),
					(int)(p.y-strbounds.getMinY()));
		}
	}

	/** Draws a single line of text using an image map as font;
	* text alignment is always top left.
	* @param imgmap  name of image map
	* @param char_offset  ASCII code of first image of image map
	* @param spacing  number of pixels to skip between letters
	*/
	public void drawImageString(String string, int x, int y,
	String imgmap, int char_offset, int spacing) {
		JGCanvas.JGImageMap map = (JGCanvas.JGImageMap)
			canvas.imagemaps.get(imgmap);
		if (map==null) return; /* not available yet */
		//Image img = map.getScaledImage();
		for (int i=0; i<string.length(); i++) {
			int imgnr = -char_offset+string.charAt(i);
			//Point coord = map.getImageCoord(imgnr);
			String lettername = imgmap+"#"+string.charAt(i);
			//System.out.println(lettername);
			Image letter = canvas.getImage(lettername);
			//System.out.println("N"+(letter==null));
			if (letter==null) {
				canvas.defineImage(lettername, "FONT", 0,
					canvas.getSubImage(imgmap,imgnr),
					true, "-", 0,0,0,0);
			}
			canvas.drawImage(buf_gfx, x,y,lettername);
			//Point scaledtl = canvas.scalePos(x, y);
			//Point scaledbr = canvas.scalePos(x+map.tilex, y+map.tiley);
			//buf_gfx.drawImage(map.img,
			//	scaledtl.x, scaledtl.y, scaledbr.x, scaledbr.y,
			//	coord.x, coord.y, coord.x+map.tilex, coord.y+map.tiley, null);
			//buf_gfx.drawImage(map.img,
			//	x, y, x+map.tilex, y+map.tiley,
			//	coord.x, coord.y, coord.x+map.tilex, coord.y+map.tiley, null);
			x += map.tilex + spacing;
		}
	}


	/** Get the key status of the given key. */
	public boolean getKey(int key) { return canvas.keymap[key]; }
	/** Set the key status of a key to released. */
	public void clearKey(int key) { canvas.keymap[key]=false; }
	/** Set the key status of a key to pressed. */
	public void setKey(int key) { canvas.keymap[key]=true; }

	/* image methods */

	/** Define image map, a large image containing a number of smaller images
	* to use for sprites or fonts.  The images must be in a regularly spaced
	* matrix.  One may define multiple image maps with the same image but
	* different matrix specs.
	* @param mapname  id of image map
	* @param imgurl  url/path of map image
	* @param xofs  x offset of first image 
	* @param yofs  y offset of first image
	* @param tilex  width of an image
	* @param tiley  height of an image
	* @param skipx  nr of pixels to skip between successive images
	* @param skipy  nr of pixels to skip between successive images vertically.
	*/
	public void defineImageMap(String mapname, String imgurl,
		int xofs,int yofs, int tilex,int tiley, int skipx,int skipy) {
		canvas.defineImageMap(mapname,imgurl, xofs,yofs, tilex,tiley,
			skipx,skipy);
	}

	/** Define new sprite/tile image. 
	* @param imgname  image id
	* @param tilename  tile id (1-4 characters)
	* @param collisionid  cid to use for tile collision matching
	* @param imgfile  filespec relative to baseurl; "null" means no file
	* @param is_transparent  signal transparency for tile operations
	* @param top  collision bounding box dimensions
	* @param left  collision bounding box dimensions
	* @param width  collision bounding box dimensions
	* @param height  collision bounding box dimensions
	*/
	public void defineImage(String imgname, String tilename, int collisionid,
	String imgfile, boolean is_transparent, String img_op,
	int top,int left, int width,int height) {
		canvas.defineImage(imgname,tilename,collisionid,imgfile,
			is_transparent, img_op, top,left,width,height);
	}

	/** Define new image from map.
	* @param imgname  image id
	* @param tilename  tile id (1-4 characters)
	* @param collisionid  cid to use for tile collision matching
	* @param imgmap  id of image map
	* @param mapidx  index of image in map, 0=first
	* @param is_transparent  signal transparency for tile operations
	* @param top  collision bounding box dimensions
	* @param left  collision bounding box dimensions
	* @param width  collision bounding box dimensions
	* @param height  collision bounding box dimensions
	*/
	public void defineImage(String imgname, String tilename, int collisionid,
	String imgmap, int mapidx, boolean is_transparent, String img_op,
	int top,int left, int width,int height) {
		canvas.defineImage(imgname,tilename,collisionid,
			canvas.getSubImage(imgmap, mapidx),
			is_transparent, img_op, top,left,width,height);
	}

	/** Load a set of imagemap, image, and animation definitions from a file.
	* The file contains one image / imagemap / animation definition
	* on each line, with the fields separated by one or more tabs.  Lines not
	* matching the required # of fields are ignored.
	* The fields have the same order as in defineImage, defineImageMap,
	* defineAnimation. */
	public void defineImages(String filename) {
		int lnr=1;
		try {
			BufferedReader in = new BufferedReader( new InputStreamReader(
				getClass().getClassLoader().getResourceAsStream( filename ) ));
			String line;
			String [] fields = new String [14];
			while ( (line = in.readLine()) != null) {
				int i=0;
				StringTokenizer toker = new StringTokenizer(line,"\t");
				while (toker.hasMoreTokens()) {
					fields[i++] = toker.nextToken();
				}
				if (i==8) {
					canvas.defineImageMap(
						fields[0], fields[1],
						Integer.parseInt(fields[2]),
						Integer.parseInt(fields[3]),
						Integer.parseInt(fields[4]),
						Integer.parseInt(fields[5]),
						Integer.parseInt(fields[6]),
						Integer.parseInt(fields[7]) );
				} else if (i==10) {
					canvas.defineImage(fields[0],fields[1],
						Integer.parseInt(fields[2]),
						fields[3],
						fields[4].equals("true"),
						fields[5],
						Integer.parseInt(fields[6]),
						Integer.parseInt(fields[7]),
						Integer.parseInt(fields[8]),
						Integer.parseInt(fields[9])  );
				} else if (i==11) {
					canvas.defineImage(fields[0],fields[1],
						Integer.parseInt(fields[2]),
						canvas.getSubImage(fields[3],
							Integer.parseInt(fields[4]) ),
						fields[5].equals("true"),
						fields[6],
						Integer.parseInt(fields[7]),
						Integer.parseInt(fields[8]),
						Integer.parseInt(fields[9]),
						Integer.parseInt(fields[10])  );
				} else if (i==3) {
					defineAnimation(fields[0], splitList(fields[1]), 
						Double.parseDouble(fields[2])  );
				} else if (i==4) {
					defineAnimation(fields[0], splitList(fields[1]),
						Double.parseDouble(fields[2]),
						Integer.parseInt(fields[3])  );
				} else if (i==5) {
					defineAnimation(fields[0], splitList(fields[1]),
						Double.parseDouble(fields[2]),
						Integer.parseInt(fields[3]),
						fields[4].equals("true")   );
				}
				lnr++;
			}
		} catch (Exception e) {
			System.err.println("Error in "+filename+" line "+lnr+":");
			e.printStackTrace();
			if (!i_am_applet) System.exit(0);
		}
		
	}

	private String [] splitList(String liststr) {
		Vector list = new Vector(10,20);
		for (StringTokenizer toker=new StringTokenizer(liststr,";");
		toker.hasMoreTokens(); ) {
			list.add(toker.nextToken());
		}
		String [] list_arr = new String [list.size()];
		int i=0;
		for (Enumeration e=list.elements(); e.hasMoreElements(); ) {
			list_arr[i] = (String) e.nextElement();
			i++;
		}
		return list_arr;
	}


	/** Gets the collision bounding box of an image. */
	public Rectangle getImageBBox(String imgname) {
		return canvas.getImageBBox(imgname);
	}

	/* background methods */

	/** Set dimensions of background. This function has been disabled as
	 * public function.  It is for defining scrolling virtual screens.
	* @param tilex x size of tile
	* @param tiley y size of tile
	* @param virtualx  number of tiles in x direction
	* @param virtualy  number of tiles in y direction
	* @param filltile  tilename of tile to use as fill pattern
	*/
	//public void setBGDim(int tilex,int tiley, int virtualx,int virtualy,
	//String filltile) {
	//	canvas.setBGDim(tilex,tiley,virtualx,virtualy,filltile);
	//}

	/** Fill the background with the given tile.
	* @param filltilename null means use background colour */
	public void clearBG(String filltilename) {
		canvas.clearBG(filltilename);
	}

	/** Define a single tile.
	*/
	public void setBGTile(int x,int y,String tilename) {
		canvas.setBGTile(x,y,tilename);
	}

	/** Set a block of tiles according to the chars in the nxm char array
	 * tilemap.  
	 */
	public void setBGTiles(int xofs,int yofs,String [] tilemap) {
		for (int y=0; y<tilemap.length; y++) {
			for (int x=0; x<tilemap[y].length(); x++) {
				canvas.setBGTile(x+xofs,y+yofs,
					new String(tilemap[y].substring(x,x+1)) );
			}
		}
	}

	/** Set a block of tiles according to the tile names in the nxm element
	* array tilemap.  Tile names in each line are separated by spaces.
	 */
	public void setBGTilesMulti(int xofs,int yofs,String [] tilemap) {
		for (int y=0; y<tilemap.length; y++) {
			StringTokenizer toker = new StringTokenizer(tilemap[y]," ");
			int x=0;
			while (toker.hasMoreTokens()) {
				canvas.setBGTile(x+xofs,y+yofs, toker.nextToken());
				x++;
			}
		}
	}

	/** Get collision Id of the tile at given pixel coordinates.
	*/
	public int getBGTileCidAtCoord(double x,double y) {
		int xidx = (int)x / canvas.tilex;
		int yidx = (int)y / canvas.tiley;
		return canvas.getBGTileCid(xidx,yidx);
	}
	/** Get cid of tile at given tile index position. */
	public int getBGTileCid(int xidx,int yidx) {
		return canvas.getBGTileCid(xidx,yidx);
	}
	/** Get the tile cid of the point that is (xofs,yofs) from the tile index
	 * coordinate center. */
	public int getBGTileCid(Point center, int xofs, int yofs) {
		return canvas.getBGTileCid(center.x+xofs, center.y+yofs);
	}

	/** Get collision Id of the tile at given pixel coordinates.
	*/
	public String getBGTileStringAtCoord(double x,double y) {
		int xidx = (int)x / canvas.tilex;
		int yidx = (int)y / canvas.tiley;
		return canvas.getBGTileString(xidx,yidx);
	}
	/** get string id of tile at given index position */
	public String getBGTileString(int xidx,int yidx) {
		return canvas.getBGTileString(xidx,yidx);
	}
	/** Get the tile string of the point that is (xofs,yofs) from the tile
	* index coordinate center. */
	public String getBGTileString(Point center, int xofs, int yofs) {
		return canvas.getBGTileString(center.x+xofs, center.y+yofs);
	}


	/** Get the OR of the cids at the tile indexes given by tiler */
	public int getBGTileCid(Rectangle tiler) {
		return canvas.getBGTileCid(tiler);
	}

	/** Get object if it exists, null if not.
	*/
	public boolean existsObject(String index) {
		return canvas.existsObject(index);
	}

	/** Get object if it exists, null if not.
	*/
	public JGObject getObject(String index) {
		return canvas.getObject(index);
	}


	/** Remove object or mark object for removal. */
	protected void removeObject(JGObject obj) {
		if (in_parallel_upd) { // queue remove
			canvas.markRemoveObject(obj);
		} else { // do remove immediately
			canvas.removeObject(obj);
		}
	}

	/** Remove all objects which have the given name prefix and/or match the
	* given cidmask.  The actual matching and removal is done after the
	* current moveObjects or check*Collision ends, or immediately if done
	* from within the main doFrame loop.  It also removes any matching
	* objects which are pending to be added the next frame.
	* @param cidmask collision id mask, 0 means ignore
	* @param prefix  ID prefix, null means ignore  */
	public void removeObjects(String prefix,int cidmask) {
		if (in_parallel_upd) {
			canvas.markRemoveObjects(prefix,cidmask);
		} else {
			canvas.removeObjects(prefix,cidmask);
		}
	}


	/** Count how many objects there are with both the given name prefix and
	* have colid&amp;cidmask != 0.  Either criterion can be left out.
	* @param cidmask collision id mask, 0 means ignore
	* @param prefix  ID prefix, null means ignore  */
	public int countObjects(String prefix,int cidmask) {
		int nr_obj=0;
		for (Enumeration e=canvas.objects.keys(); e.hasMoreElements();) {
			String name = (String) e.nextElement();
			if (prefix==null || name.startsWith(prefix)) {
				if (cidmask==0 || (getObject(name).colid&cidmask)!=0) {
					nr_obj++;
				}
			}
		}
		return nr_obj;
	}


	/** Call the move() methods of those objects matching the given name
	 * prefix and collision id mask.
	* @param cidmask collision id mask, 0 means ignore
	* @param prefix  ID prefix, null means ignore  */
	public void moveObjects(String prefix, int cidmask) {
		if (in_parallel_upd) throw new Error("Recursive call");
		in_parallel_upd=true;
		canvas.moveObjects(prefix,cidmask);
		in_parallel_upd=false;
	}

	/** Call all move() methods of all registered objects. */
	public void moveObjects() {
		if (in_parallel_upd) throw new Error("Recursive call");
		in_parallel_upd=true;
		canvas.moveObjects(null,0); 
		in_parallel_upd=false;
	}

	/** Calls all colliders of objects that match dstid that collide with
	* objects that match srcid.
	*/
	public void checkCollision(int srcid,int dstid) {
		if (in_parallel_upd) throw new Error("Recursive call");
		in_parallel_upd=true;
		canvas.checkCollision(srcid,dstid);
		in_parallel_upd=false;
	}


	/** Check collision of tiles within given rectangle, return the OR of all
	* cids found.
	*/
	public int checkBGCollision(Rectangle r) {
		return canvas.checkBGCollision(r);
	}

	/** Calls all bg colliders of objects that match objid that collide with
	* tiles that match tileid.
	*/
	public void checkBGCollision(int tileid,int objid) {
		if (in_parallel_upd) throw new Error("Recursive call");
		in_parallel_upd=true;
		canvas.checkBGCollision(tileid,objid);
		in_parallel_upd=false;
	}


	/* computation */

	/** A Boolean AND shorthand to use for collision;
	* returns (value&amp;mask) != 0. */
	public boolean and(int value, int mask) {
		return (value&mask) != 0;
	}

	public double random(double min, double max) {
		return min + Math.random()*(max-min);
	}

	public double random(double min, double max, double interval) {
		int steps = (int)Math.floor((max-min)/interval);
		return min + ((int) (Math.random()*(steps+0.95)))*interval;
	}

	/** Get tile index range of all tiles overlapping given rectangle of pixel
	 * coordinates. 
	 * @return tile indices */
	public Rectangle getTileSpan(Rectangle r) {
		return canvas.getTileSpan(r);
	}

	/** Get tile index of the tile the coordinate is on.*/
	public Point getTileIndex(double x, double y) {
		return new Point(
			(int)Math.floor( x / (double)canvas.tilex ),
			(int)Math.floor( y / (double)canvas.tiley )  );
	}

	/** Snap to grid, double version. Epsilon is added to the gridsnap value,
	* so that isXAligned(x,margin) always implies that snapToGridX(x,margin)
	* will snap.
	* @param x  position to snap
	* @param gridsnapx  snap margin, 0.0 means no snap   */
	public double snapToGridX(double x, double gridsnapx) {
		if (gridsnapx == 0.0) return x;
		int xaligned = canvas.tilex*(int)
			Math.floor(((x + canvas.tilex/2.0) / (double)canvas.tilex));
		double gridofsx = Math.abs(x - xaligned);
		if (gridofsx <= gridsnapx+0.01) return xaligned;
		return x;
	}

	/** Snap to grid, double version. Epsilon is added to the gridsnap value,
	* so that isYAligned(y,margin) always implies that snapToGridY(y,margin)
	* will snap.
	* @param y  position to snap
	* @param gridsnapy  snap margin, 0.0 means no snap   */
	/** Snap to grid, double version. Epsilon is added to the gridsnap value.
	* 0.0 means no snap */
	public double snapToGridY(double y, double gridsnapy) {
		if (gridsnapy == 0.0) return y;
		int yaligned = canvas.tiley*(int)
			Math.floor(((y + canvas.tiley/2.0) / (double)canvas.tiley));
		double gridofsy = Math.abs(y - yaligned);
		if (gridofsy <= gridsnapy+0.01) return yaligned;
		return y;
	}

	/** Snap p to grid in case p is close enough to the grid lines. Note: this
	 * function only handles integers so it should not be used to snap an
	 * object position.  */
	public void snapToGrid(Point p,int gridsnapx,int gridsnapy) {
		if (gridsnapx==0 && gridsnapy==0) return;
		int xaligned = canvas.tilex*(int)
			Math.floor(((p.x + canvas.tilex/2.0) / (double)canvas.tilex));
		int yaligned = canvas.tiley*(int)
			Math.floor(((p.y + canvas.tiley/2.0) / (double)canvas.tiley));
		int gridofsx = Math.abs(p.x - xaligned);
		int gridofsy = Math.abs(p.y - yaligned);
		if (gridofsx <= gridsnapx) p.x = xaligned;
		if (gridofsy <= gridsnapy) p.y = yaligned;
	}

	/** Returns true if x falls within margin of the tile snap grid. Epsilon
	 * is added to the margin, so that isXAligned(1.000..., 1.000...)
	 * always returns true. */
	public boolean isXAligned(double x,double margin) {
		int xaligned = canvas.tilex*(int)(((int)x + canvas.tilex/2)
			/ canvas.tilex);
		return Math.abs(x - xaligned) <= margin+0.001;
	}

	/** Returns true if y falls within margin of the tile snap grid. Epsilon
	 * is added to the margin, so that isYAligned(1.000..., 1.000...)
	 * always returns true. */
	public boolean isYAligned(double y,double margin) {
		int yaligned = canvas.tiley*(int)(((int)y + canvas.tiley/2)
			/ canvas.tiley);
		return Math.abs(y - yaligned) <= margin+0.001;
	}


	/* listening */

	class JGListener implements KeyListener, WindowListener {
		public JGListener () {}
		/* Standard Wimp event handlers */
		public void keyPressed(KeyEvent e) {
			int keychar = e.getKeyChar();
			int keycode = e.getKeyCode();
			if (keycode>=0 && keycode < 256) {
				canvas.keymap[keycode]=true;
			}
			//System.out.println(e+" keychar"+e.getKeyChar());
		}
		public void keyReleased(KeyEvent e) {
			int keychar = e.getKeyChar();
			int keycode = e.getKeyCode();
			if (keycode>=0 && keycode < 256) {
				canvas.keymap[keycode]=false;
			}
		}
		public void keyTyped(KeyEvent e) {
			/* shift escape = exit */
			if (e.isShiftDown() && e.getKeyChar() == 27 && !i_am_applet) {
				System.exit(0);
			}
		}
		/* WindowListener handlers */

		public void windowActivated(WindowEvent e) {}
		public void windowClosed(WindowEvent e) {System.out.println("Closed");}
		public void windowClosing(WindowEvent e) {
			System.out.println("Window closed; exiting.");
			closeWindow();
		}
		public void windowDeactivated(WindowEvent e) {}
		public void windowDeiconified(WindowEvent e) {}
		public void windowIconified(WindowEvent e) {}
		public void windowOpened(WindowEvent e) {}

	}


	/** Engine thread, executing game action. */
	class JGEngineThread implements Runnable {
		public JGEngineThread () {}
		public void run() { try {
			initGame();
			canvas.setInitialised();
			target_time = System.currentTimeMillis()+(long)(1000.0/fps);
			while (!is_exited) {
				long cur_time = System.currentTimeMillis();
				if (!running) {
					// wait in portions of 1/2 sec until running is set;
					// reset target time
					Thread.sleep(500);
					target_time = cur_time+(long)(1000.0/fps);
				} else if (cur_time < target_time+(long)(330.0/fps)) {
					// we lag behind less than 1/3 frame -> do full frame
					if (cur_time+3 < target_time)
						//we even have some time left -> sleep it away
						Thread.sleep(target_time-cur_time);
					doFrameAll();
					canvas.repaint();
					target_time += (1000.0/fps);
				} else if (cur_time >
				target_time + (long)(1000.0*maxframeskip/fps)) {
					// we lag behind more than the max # frames ->
					// draw full frame and reset target time
					doFrameAll();
					canvas.repaint();
					target_time=cur_time + (long)(1000.0/fps);
				} else {
					// we lag behind a little -> frame skip
					doFrameAll();
					target_time += (long)(1000.0/fps);
				}
			}
		} catch (InterruptedException e) {
			/* exit thread when interrupted */
			System.out.println("JGame thread exited.");
		} }
	}


	/** create window to simulate applet window.
	*/
	private void createWindow() {
		/* create frame to `emulate' an applet's frame */
		my_win = new Frame();
		/*{
			protected void processKeyEvent(KeyEvent e) {
				System.out.println(e+" keychar"+e.getKeyChar());
			}
			protected void  processFocusEvent(FocusEvent e) {
				System.out.println("focus"+e);
			}
			public boolean isFocusTraversable() { return true; }
		};*/
		// ensure no margins around canvas
		my_win.setLayout(new FlowLayout(FlowLayout.LEADING,0,0));
		// setting the size of the canvas or applet has no effect
		// we need to add the height of the title bar to the height
		// XXX can't figure out how to get the title bar height
		// 24 is the empirically determined height in WinXP
		// 32 enables us to have the whole window on-screen
		my_win.setSize(width, height+32);
		my_win.add(this);
		my_win.addWindowListener(listener);
		//my_win.addKeyListener(listener); this one worked in both 1.2 and 1.4

	}
	private void closeWindow() {
		my_win.setVisible(false);
		System.exit(0);
	}


}

