package jgame;
import java.awt.*;
import java.awt.image.ImageObserver;
import java.io.Serializable;
import java.net.*;

/** Superclass for game objects, override to define animated game objects.
 * When an object is created, it is automatically registered with the
 * currently running engine.  The object will become active only after the
 * frame ends.  The object is managed by the engine, which will display it and
 * call the move and hit methods when appropriate.  Call remove() to remove
 * the object.  It will be removed after the frame ends.  Use isAlive()
 * to see if the object has been removed or not.

 * <p>
 * Each object corresponds to one image.  If you want multi-image objects,
 * simply use multiple objects and co-reference them using regular references
 * or using JGEngine's getObject().  The object's appearance can be changed
 * using setImage or any of the animation functions.  

 * <p>
 * Objects have a pointer to the engine by which they are managed (eng).  This
 * can be used to call the various useful methods in the engine.
 * Alternatively, the objects can be made inner classes of your JGEngine
 * class, so that they have direct access to all JGEngine methods. 

 * <p> The object remembers some of the state of the previous frame (in
 * particular the previous position and bounding boxes), so that corrective
 * action can be taken after something special happened (such as bumping into
 * a wall).  Alternatively tile reading functions can be used, such as
 * BGTileCid().

 * Objects have a direction and speed built in.  After their move() method
 * finished, their x and y position are incremented with the given
 * speed/direction.  Speed may be used as absolute value (leave the direction
 * at 1), or as a value relative to the direction (the movement is speed*dir).

 */
public class JGObject {
	static int next_id = 0; 
	/* global which might be accessed concurrently */
	static JGCanvas default_canvas=null;
	protected static JGEngine default_engine=null;
	/** @return true if success; false if other engine already running */
	protected static boolean setEngine(JGEngine engine,JGCanvas canvas) {
		if (default_engine!=null) return false;
		default_engine=engine;
		default_canvas=canvas;
		return true;
	}

	/** Object position */
	public double x=0, y=0;
	/** Object speed; default=0 */
	public double xspeed=0, yspeed=0;
	/** Object direction, is multiplied with speed; default=1 */
	public int xdir=1,ydir=1;
	/** Collision ID */
	public int colid;
	/** Object's global identifier; may not change during the lifetime of the
	 * object. */
	protected String name;
	/** Set number of move() steps before object removes itself, -1 (default)
	 * is never; -2 means expire when off-screen. */
	public int expiry=-1;
	/** Get object's ID */
	public String getName() { return name; }

	boolean is_alive=true;

	String imgname=null;
	JGEngine.Animation anim=null; /* will update imgname if set */
	String animid=null;
	/** cached value: has to be recomputed when image changes; simply set to
	 * null to do this. */
	Rectangle imgbbox=null;
	/** tile bbox is the bbox with offset 0; we have to add the current
	 * coordinate to obtain the actual tile bbox.  Set to null to use regular
	 * bbox instead. */
	Rectangle tilebbox=null;
	/** The bbox that should override the image bbox if not null. */
	Rectangle bbox=null;
	private JGCanvas canvas;

	/** You can use this to call methods in the object's engine.  Even handier
	 * is to have the objects as inner class of the engine. */
	public JGEngine eng;

	/* dimensions of last time drawn  */
	double lastx=0, lasty=0;
	Rectangle lastbbox=null;
	Rectangle lasttilebbox=null; /* actual coordinates */

	private void initObject(JGEngine engine,JGCanvas canvas,
	String name,int collisionid) {
		this.eng=engine;
		this.canvas=canvas;
		this.name=name;
		colid=collisionid;
		canvas.markAddObject(this);
	}

	/** Clear tile bbox definition so that we use the regular bbox again. */
	public void clearTileBBox() { tilebbox=null; }

	public void setTileBBox(int x,int y, int width,int height) {
		tilebbox = new Rectangle(x,y,width,height);
	}

	/** Set bbox definition to override the image bbox.  */
	public void setBBox(int x,int y, int width,int height) {
		bbox = new Rectangle(x,y,width,height);
	}
	/** Clear bbox definition so that we use the image bbox again. */
	public void clearBBox() { bbox=null; }


	///** Set object's tile span by defining the number of tiles and the margin
	// * by which the object's position is snapped for determining the object's
	// * top left position.  */
	//public void setTileSpan(int xtiles,int ytiles,int gridsnapx,int gridsnapy){
	//	this.xtiles=xtiles;
	//	this.ytiles=ytiles;
	//	this.gridsnapx=gridsnapx;
	//	this.gridsnapy=gridsnapy;
	//}

	public void setPos(double x,double y) {
		this.x=x;
		this.y=y;
	}

	/** Set absolute speed.  Set xdir, ydir to the sign of the supplied speed,
	 * and xspeed and yspeed to the absolute value of the supplied speed.
	 * Passing a value of exactly 0.0 sets the dir to 0. */
	public void setSpeedAbs(double xspeed, double yspeed) {
		if (xspeed < 0.0) {
			xdir = -1;
			this.xspeed = -xspeed;
		} else if (xspeed == 0.0) {
			xdir = 0;
			this.xspeed = 0;
		} else {
			xdir = 1;
			this.xspeed = xspeed;
		}
		if (yspeed < 0.0) {
			ydir = -1;
			this.yspeed = -yspeed;
		} else if (yspeed == 0.0) {
			ydir = 0;
			this.yspeed = 0;
		} else {
			ydir = 1;
			this.yspeed = yspeed;
		}
	}

	/** Set speed and direction in one go. */
	public void setDirSpeed(int xdir,int ydir, double xspeed, double yspeed) {
		this.xdir=xdir;
		this.ydir=ydir;
		this.xspeed=xspeed;
		this.yspeed=yspeed;
	}

	/** Set speed and direction in one go. */
	public void setDirSpeed(int xdir,int ydir, double speed) {
		this.xdir=xdir;
		this.ydir=ydir;
		this.xspeed=speed;
		this.yspeed=speed;
	}

	/** Set relative speed; the values are copied into xspeed,yspeed. */
	public void setSpeed(double xspeed, double yspeed) {
		this.xspeed=xspeed;
		this.yspeed=yspeed;
	}

	/** Set relative speed; the value is copied into xspeed,yspeed. */
	public void setSpeed(double speed) {
		this.xspeed=speed;
		this.yspeed=speed;
	}

	/** Set direction. */
	public void setDir(int xdir, int ydir) {
		this.xdir=xdir;
		this.ydir=ydir;
	}

	/** Set ID of animation or image to display.  First, look for an animation
	 * with the given ID, and setAnim if found.  Otherwise, look for an image
	 * with the given ID, and setImage if found.  Passing null clears the
	 * image and stops the animation.  */

	public void setGraphic(String gfxname) {
		if (gfxname==null) {
			setImage(gfxname);
		} else {
			JGEngine.Animation newanim = eng.getAnimation(gfxname);
			if (newanim!=null) {
				setAnim(gfxname);
			} else {
				setImage(gfxname);
			}
		}
	}

	/** Set ID of image to display; stop animation if one was running */
	public void setImage(String imgname) {
		this.imgname=imgname;
		imgbbox=null;
		stopAnim();
	}
	/** Get object's current image ID */
	public String getImage(String imgname) { return imgname; }

	/* animation */

	/** Set the animation to the given default animation definition, or leave
	 * it as it was if the anim_id is unchanged.  Subsequent changes made in
	 * the animation's parameters do not change the default animation
	 * definition. The changes will be preserved if another call to
	 * setAnimation is made with the same anim_id.  If you want to reset the
	 * animation to the original settings, use resetAnimation().
	 */
	public void setAnim(String anim_id) {
		if (animid==null || !animid.equals(anim_id)) {
			anim = eng.getAnimation(anim_id);
			if (anim==null) {
				System.out.println("Error: animation "+anim_id+" not found.");
				return;
			}
			anim = anim.copy();
			animid = anim_id;
			imgname = anim.getCurrentFrame();
		}
	}

	/** Always set the animation to the given default animation definition,
	 * resetting any changes or updates made to the animation. Subsequent
	 * changes made in the animation's parameters do not change the default
	 * animation definition.
	 */
	public void resetAnim(String anim_id) {
		anim = eng.getAnimation(anim_id).copy();
		animid = anim_id;
	}

	/** Clear the animation, the object's current image will remain. */
	public void clearAnim() { anim=null; animid=null; }

	public String getAnimId() { return animid; }

	public void setAnimIncrement(int increment) {
		if (anim!=null) anim.increment=increment;
	}
	public void setAnimSpeed(double speed) {
		if (anim!=null) anim.speed=speed;
	}
	public void setAnimPingpong(boolean pingpong) {
		if (anim!=null) anim.pingpong=pingpong;
	}
	public void startAnim() { if (anim!=null) anim.start(); }
	public void stopAnim() { if (anim!=null) anim.stop(); }

	/** Reset the animation's state to the start state. */
	public void resetAnim() { if (anim!=null) anim.reset(); }


	/** Create object.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
	}

	/** Create object with given tile bbox.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	int tilebbox_x,int tilebbox_y, int tilebbox_width,int tilebbox_height) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setTileBBox(tilebbox_x,tilebbox_y,tilebbox_width,tilebbox_height);
	}

	/** Create object with given absolute speed.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	double xspeed, double yspeed) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setSpeedAbs(xspeed,yspeed);
	}

	/** Create object with given tile bbox and absolute speed.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	int tilebbox_x,int tilebbox_y, int tilebbox_width,int tilebbox_height,
	double xspeed, double yspeed) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setTileBBox(tilebbox_x,tilebbox_y,tilebbox_width,tilebbox_height);
		setSpeedAbs(xspeed,yspeed);
	}

	/** Create object with given absolute speed, expiry.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	double xspeed, double yspeed, int expiry) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setSpeedAbs(xspeed,yspeed);
		this.expiry = expiry;
	}

	/** Create object with given tile bbox, absolute speed, expiry.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	int tilebbox_x,int tilebbox_y, int tilebbox_width,int tilebbox_height,
	double xspeed, double yspeed, int expiry) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setTileBBox(tilebbox_x,tilebbox_y,tilebbox_width,tilebbox_height);
		setSpeedAbs(xspeed,yspeed);
		this.expiry = expiry;
	}

	/** Create object with given direction/speed, expiry.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	int xdir, int ydir, double xspeed, double yspeed, int expiry) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setDirSpeed(xdir,ydir,xspeed,yspeed);
		this.expiry = expiry;
	}

	/** Create object with given tile bbox, direction/speed, expiry.
	* @param unique_id  append name with unique ID if unique_id set
	* @param gfxname  id of animation or image, null = no image */
	public JGObject (String name, boolean unique_id,
	double x,double y,int collisionid,String gfxname,
	int tilebbox_x,int tilebbox_y, int tilebbox_width,int tilebbox_height,
	int xdir, int ydir, double xspeed, double yspeed, int expiry) {
		initObject(default_engine,default_canvas,
				name + (unique_id ? ""+(next_id++) : "" ), collisionid );
		setPos(x,y);
		setGraphic(gfxname);
		setTileBBox(tilebbox_x,tilebbox_y,tilebbox_width,tilebbox_height);
		setDirSpeed(xdir,ydir,xspeed,yspeed);
		this.expiry = expiry;
	}


	/* Bounding box functions.  Return copies. May return null if
	 * image is not defined. */


	/** Get collision bounding box in pixels. Has actual coordinate offset.
	* @return copy of bbox in pixel coordinates, null if no bbox
	*/
	public Rectangle getBBox() {
		if (bbox!=null) return new Rectangle(bbox.x+(int)x, bbox.y+(int)y,
				bbox.width, bbox.height);
		if (imgbbox==null) {
			imgbbox = getImageBBox();
		}
		if (imgbbox!=null) {
			return new Rectangle(imgbbox.x+(int)x, imgbbox.y+(int)y,
				imgbbox.width, imgbbox.height );
		}
		return null;
	}
/*	Rectangle temp_bbox = new Rectangle(0,0,0,0);
	//an attempt at optimisation of new Rectangle
	public Rectangle getBBox() {
		if (bbox!=null) {
			temp_bbox.x = bbox.x+(int)x;
			temp_bbox.y = bbox.y+(int)y;
			temp_bbox.width = bbox.width;
			temp_bbox.height= bbox.height;
			return temp_bbox;
		}
		if (imgbbox==null) {
			imgbbox = getImageBBox();
		}
		if (imgbbox!=null) {
			temp_bbox.x = imgbbox.x+(int)x;
			temp_bbox.y = imgbbox.y+(int)y;
			temp_bbox.width=imgbbox.width;
			temp_bbox.height=imgbbox.height;
			return temp_bbox;
		}
		return null;
	} */

	/** Get collision bounding box in pixels of previous frame.
	* @return pixel coordinates, null if no bbox */
	public Rectangle getLastBBox() {
		if (lastbbox==null) return null;
		return new Rectangle(lastbbox);
	}

	/** Get collision bounding box in pixels for the purpose of colliding with
	 * tiles.  Bounding box has actual coordinate offset.
	* @return  pixel coordinates, null if no bbox */
	public Rectangle getTileBBox() {
		if (tilebbox==null) return getBBox();
		return new Rectangle((int)x+tilebbox.x,(int)y+tilebbox.y,
			tilebbox.width,tilebbox.height);
	}

	/** Get tile collision bounding box of previous frame.
	* @return  pixel coordinates, null if no bbox */
	public Rectangle getLastTileBBox() {
		if (lasttilebbox==null) return null;
		return new Rectangle(lasttilebbox);
	}

	/** Get collision bounding box of object's image (same as object's, only
	 * with offset (0,0)).
	* @return  pixel coordinates, null if no bbox */
	public Rectangle getImageBBox() {
		if (imgbbox==null && imgname!=null) {
			imgbbox = new Rectangle(canvas.getImageBBox(imgname));
		}
		if (imgbbox==null) return null;
		return new Rectangle(imgbbox);
	}

	/** Get x position of previous frame. */
	public double getLastX() { return lastx; }
	/** Get y position of previous frame. */
	public double getLastY() { return lasty; }


	/* snap functions */

	/** Snap object to grid using the default gridsnap margin of (xspeed/2,
	 * yspeed/2), corresponding to the default is...Aligned margin. */
	public void snapToGrid() {
		x = eng.snapToGridX(x,Math.abs(xspeed/2.0));
		y = eng.snapToGridY(y,Math.abs(yspeed/2.0));
	}

	/** Snap object to grid.
	* @param gridsnapx margin below which to snap, 0.0 is no snap
	* @param gridsnapy margin below which to snap, 0.0 is no snap */
	public void snapToGrid(double gridsnapx,double gridsnapy) {
		x = eng.snapToGridX(x,gridsnapx);
		y = eng.snapToGridY(y,gridsnapy);
	}
	///** Snap object to grid. */
	//public void snapToGrid(int gridsnapx, int gridsnapy) {
	//	Point p = new Point((int)x,(int)y);
	//	eng.snapToGrid(p,gridsnapx,gridsnapy);
	//	x = p.x;
	//	y = p.y;
	//}
	/** Snap an object's tile bbox corner to grid; floats are rounded down.
	* Snaps to bottom or right of object instead of top and left if the resp.
	* flags are set. */
	public void snapBBoxToGrid(double gridsnapx,double gridsnapy,
	boolean snap_right, boolean snap_bottom) {
		Rectangle bbox = getImageBBox();
		double bx = x + bbox.x;
		double by = y + bbox.y;
		if (snap_right)  bx += bbox.width;
		if (snap_bottom) by += bbox.height;
		double bxs = eng.snapToGridX(bx,gridsnapx);
		double bys = eng.snapToGridY(by,gridsnapy);
		bxs -= bbox.x;
		bys -= bbox.y;
		if (snap_right)  bxs -= bbox.width;
		if (snap_bottom) bys -= bbox.height;
		x = bxs;
		y = bys;
	}

	/* bg interaction */

	/** Get the tile index coordinates of the object's tile bbox.
	* @return tile index coordinates, null if no bbox */
	public Rectangle getTileSpan() {
		return canvas.getTileSpan(getTileBBox());
	}

	/** Get the tile index coordinates of the object's previous tile bbox.
	* @return tile index coordinates, null if no bbox */
	public Rectangle getLastTileSpan() {
		return canvas.getTileSpan(getLastTileBBox());
	}

	/** Get the tile indices spanning the tiles that the object has the
	* most overlap with.  If a tilebbox is defined, the size of the span is
	* always the same as that of the tilebbox.  If none is defined, the size
	* is the same as bbox / nr_tiles rounded up.
	*
	* @return  tile index coordinates, null if no tile bbox is defined */
	public Rectangle getCenterTiles() {
		Rectangle bbox = getTileBBox();
		if (bbox==null) return null;
		/* find the tile on which the center of the bounding box is located,
		 * that will be the center of our tile span. */
		Point p = new Point(bbox.x + (bbox.width/2), bbox.y + (bbox.height/2));
		p = eng.getTileIndex(p.x,p.y);
		p.x = p.x*canvas.tilex + canvas.tilex/2;
		p.y = p.y*canvas.tiley + canvas.tiley/2;
		bbox.x = canvas.tilex * ( (p.x-(bbox.width /2)) / canvas.tilex);
		bbox.y = canvas.tiley * ( (p.y-(bbox.height/2)) / canvas.tiley);
		//if (bbox.width == 16) System.out.println("span"+canvas.getTileSpan(bbox));
		//bbox.width  = (int)((bbox.width  + canvas.tilex - 1) / canvas.tilex);
		//bbox.height = (int)((bbox.height + canvas.tiley - 1) / canvas.tiley);
		return canvas.getTileSpan(bbox);
	}

	/** @return tile index coordinates, null if no bbox */
	public Rectangle getLastCenterTiles() {
		Rectangle bbox = getLastTileBBox();
		Point p = new Point(bbox.x,bbox.y);
		eng.snapToGrid(p, canvas.tilex, canvas.tiley);
		bbox.x = p.x;
		bbox.y = p.y;
		return canvas.getTileSpan(bbox);
	}

	/** Get the top left coordinate of the center tiles of the object,
	* i.e. the x and y of getCenterTiles().  
	* @return tile index coordinate, null if no bbox */
	public Point getCenterTileTopLeft() {
		Rectangle bbox = getTileBBox();
		if (bbox==null) return null;
		/* find the tile on which the center of the bounding box is located,
		 * that will be the center of our tile span. */
		Point p = new Point(bbox.x + (bbox.width/2), bbox.y + (bbox.height/2));
		p = eng.getTileIndex(p.x,p.y);
		p.x = p.x*canvas.tilex + canvas.tilex/2;
		p.y = p.y*canvas.tiley + canvas.tiley/2;
		return new Point(
				( (p.x-(bbox.width /2)) / canvas.tilex),
				( (p.y-(bbox.height/2)) / canvas.tiley)  );
	}

	/** Get the top left tile of the object's tile bbox.
	* @return tile index coordinate, null if no bbox */
	public Point getTopLeftTile() {
		Rectangle r = getTileSpan();
		if (r==null) return null;
		return eng.getTileIndex(r.x,r.y);
	}

	/** Returns true if x is distance xspeed/2 away from being grid aligned.
	 * If an object moves at its current xspeed, this method will always
	 * return true when the object crosses the tile alignment line. */
	public boolean isXAligned() {
		return eng.isXAligned(x,Math.abs(xspeed*0.5));
	}

	/** Returns true if y is distance yspeed/2 away from being grid aligned.
	 * If an object moves at its current yspeed, this method will always
	 * return true when the object crosses the tile alignment line. */
	public boolean isYAligned() {
		return eng.isYAligned(y,Math.abs(yspeed*0.5));
	}

	/** Returns true if x is within margin of being tile grid aligned. */
	public boolean isXAligned(double margin) {
		return eng.isXAligned(x,margin);
	}

	/** Returns true if y is within margin of being tile grid aligned. */
	public boolean isYAligned(double margin) {
		return eng.isYAligned(y,margin);
	}

	/** Returns true if the left of the object's tile bbox is within margin of
	* being tile grid aligned. */
	public boolean isLeftAligned(double margin) {
		Rectangle bbox = getTileBBox();
		if (bbox!=null) {
			return eng.isXAligned(bbox.x,margin);
		} else {
			return eng.isXAligned(x,margin);
		}
	}

	/** Returns true if the top of the object's tile bbox is within margin of
	* being tile grid aligned. */
	public boolean isTopAligned(double margin) {
		Rectangle bbox = getTileBBox();
		if (bbox!=null) {
			return eng.isYAligned(bbox.y,margin);
		} else {
			return eng.isYAligned(y,margin);
		}
	}

	/** Returns true if the right of the object's tile bbox is within margin of
	* being tile grid aligned. */
	public boolean isRightAligned(double margin) {
		Rectangle bbox = getTileBBox();
		if (bbox!=null) {
			return eng.isXAligned(bbox.x+bbox.width,margin);
		} else {
			return eng.isXAligned(x,margin);
		}
	}

	/** Returns true if the bottom of the object's tile bbox is within margin of
	* being tile grid aligned. */
	public boolean isBottomAligned(double margin) {
		Rectangle bbox = getTileBBox();
		if (bbox!=null) {
			return eng.isYAligned(bbox.y+bbox.height,margin);
		} else {
			return eng.isYAligned(y,margin);
		}
	}

	/* something odd going on here: it fails to find the checkBGCollision in
	 * Engine when I define this one in Object
	public int checkBGCollision() {
		return eng.checkBGCollision(getTileBBox());
	}*/

	/** Margin is the margin beyond which the object is considered offscreen.
	* The tile bounding box is used as the object's size, if there is none, we
	* use a null bounding box.
	* @param marginx pixel margin, may be negative
	* @param marginy pixel margin, may be negative
	*/
	public boolean isOnScreen(int marginx,int marginy) {
		Rectangle bbox = getTileBBox();
		if (bbox==null) bbox = new Rectangle(0,0,0,0);
		if (bbox.x + bbox.width  < -marginx) return false;
		if (bbox.y + bbox.height < -marginy) return false;
		if (bbox.x > eng.pfWidth() + marginx) return false;
		if (bbox.y > eng.pfHeight() + marginy) return false;
		return true;
	}


	/** prepare object for beginning of frame; animate */
	protected final void prepareForFrame() {
		if (anim!=null) {
			imgname = anim.animate();
			imgbbox=null;
		}
	}
	/** a new frame has just been drawn; make snapshot of state */
	protected final void frameFinished() {
		lastx=x;
		lasty=y;
		lastbbox = getBBox();
		lasttilebbox = getTileBBox();
	}

	/** Check if object is still active, or has already been removed. */
	public boolean isAlive() { return is_alive; }

	/** Signal to object that remove is done. */
	protected void removeDone() { is_alive=false; }

	/** Mark object for removal.
	* The object will be removed at the end of this
	* moveObjects, checkCollision, checkBGCollision, or frame. */
	public void remove() { if (isAlive()) eng.removeObject(this); }

	/** Override to implement automatic move; default is do nothing.
	*/
	public void move() {}

	/** Override to handle collision; default is do nothing.
	*/
	public void hit(JGObject obj) {}

	/** Override to handle tile collision; default is do nothing.
	*/
	public void hit_bg(int tilecid) {}
	
	/** Override to handle tile collision; default is do nothing.
	* This method is always called when hit_bg() is also called, only, extra
	* information is passed that may be useful.
	*/
	public void hit_bg(int tilecid,int tx,int ty, int txsize,int tysize) {}
}

