package jgame;

/** A timer that generates a callback after a certain number of frames.  It
 * can conveniently be used as a "one liner class", i.e. as an inner class
 * within a context where it can set a variable or call a method. i.e.
 *<code> new JGTimer (10,true) { alarm() { doSomething(); } };</code>

 * <P>Timers are updated just before the beginning of a frame.  Any objects
 * they add or delete are immediately updated before the frame starts.

*/
public abstract class JGTimer {
	int frames=0, frames_left=0;
	public boolean running=true;
	public boolean one_shot;

	/** Create timer; the timer may be one-shot (it runs only once, then
	 * triggers the alarm and removes itself), or continuous (it continues
	 * running and triggering the alarm)
	* @param frames_to_alarm  0 = callback just before next frame.
	* @param one_shot  true = run only once, false = run repeatedly
	*/
	public JGTimer(int frames_to_alarm, boolean one_shot) {
		set(frames_to_alarm,one_shot);
		JGObject.default_engine.registerTimer(this);
	}

	public void set(int frames_to_alarm, boolean one_shot) {
		frames=frames_to_alarm;
		frames_left=frames;
		this.one_shot = one_shot;
	}

	/** Tick function, as called by the engine. Returns true when the timer
	 * has to be removed, */
	protected boolean tick() {
		if (running) {
			if (frames_left <= 0) {
				frames_left=frames;
				alarm();
				if (one_shot) return true;
			} else {
				frames_left--;
			}
		}
		return false;
	}

	/** define your alarm action here. */
	abstract public void alarm();
}

